/++
 + テンプレートエンジン スイーツ(笑) Version: 0.4.1
 +
 + 拡張機能モジュール
 +
 + Copyright(C) 2009-2011 Takuya Kurosawa
 +/
module jp.ku6.sweets.extension;

import std.datetime;
import std.file;
import jp.ku6.sweets.core;

/++
 + AutoUpdatableSweetsに使用するテンプレート
 +/
interface AutoUpdatableSweetsTemplate
{
	/++
	 + リビジョン番号を取得
	 +
	 + elementsで取得出来るデータの更新チェックの為
	 + 更新毎にインクリメントする番号
	 +
	 + Returns:
	 +  リビジョン番号 1から開始する
	 +/
	uint revision();

	/++
	 + データを取得
	 +
	 + Returns:
	 +  テキストテンプレートを解析した結果データ
	 +/
	SweetsElement[] elements();
}


/++
 + 自動更新機能付きSweets
 +/
class AutoUpdatableSweets : Sweets
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  tpl = テンプレート
	 +/
	this(AutoUpdatableSweetsTemplate tpl) 
	{
		tpl_ = tpl;
	}

	override
	void assign(string id, SweetsProcessor processor)
	{
		// 要素が更新される場合があるのでdelegateはインデックスを作っておく
		processors_[id] = processor;

		// 要素に処理を割り当て
		super.assign(id, processor);
	}

	override
	void print(SweetsPrinter printer)
	{
		if (tpl_.revision() != revision_)
		{
			// リビジョンを更新
			revision_ = tpl_.revision();

			// 要素を取得
			SweetsElement[] elements = tpl_.elements();

			children_ = new SweetsNode[elements.length];
			for (uint i; i < children_.length; ++i)
			{
				children_[i] = new SweetsNode(elements[i]);
				reindexElements(children_[i]);
			}

			// 要素に対する操作を再度割り付ける
			if (processors_.length > 0)
			{
				foreach (id, processor; processors_)
				{
					SweetsNode[]* nodes = id in nodeIndex_;
					if (nodes !is null)
					{
						foreach (node; *nodes)
							node.processor = processor;
					}
				}
			}
		}

		foreach (child; children_)
			child.print(printer);
	}

protected:

	// 元となるテンプレートデータ
	AutoUpdatableSweetsTemplate tpl_;

	// 更新を関しするためのリビジョン
	uint revision_;

	// IDに対する処理のインデックス
	SweetsProcessor[string] processors_;

	/+
	 + コンストラクタ
	 + 
	 + 継承先で空のオブジェクトを生成するためのコンストラクタ
	 +/
	this()
	{
	}
}

/++
 + ファイル更新を監視して自動更新するSweetsTemplate
 +/
class FileAutoReloadableSweetsTemplate : AutoUpdatableSweetsTemplate
{
	/++
	 + インスタンスを取得
	 +
	 + ファイルパスを指定し、そのファイルに対応したインスタンスを取得する
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +
	 + Returns:
	 +  インスタンス
	 +/
	synchronized static
	FileAutoReloadableSweetsTemplate instance(string filePath)
	{
		File* file = filePath in files_;
		if (file is null)
		{
			File f = new File(filePath);
			files_[filePath] = f;
			return new FileAutoReloadableSweetsTemplate(f);
		}
		else
			return new FileAutoReloadableSweetsTemplate(*file);
	}

	override
	uint revision()
	{
		synchronized (file_)
		{
			file_.refresh();
			elements_ = file_.elements;
			revision_ = file_.revision;
		}

		return revision_;
	}

	override
	SweetsElement[] elements()
	{
		return elements_;
	}

protected:

	shared static
	class File
	{
		// パス
		string path;

		// 更新時間
		long time;

		// リビジョン
		uint revision;

		// 解析結果の要素データ
		SweetsElement[] elements;

		/+
		 + コンストラクタ
		 +
		 + Params:
		 +  p = ファイルパス
		 +/
		this(string p)
		{
			path = p;
			revision = 1;
		}

		/+
		 + 読み込み済みのデータとファイルのタイムスタンプを比較し、更新されていたときに再読み込みを行う
		 +/
		void refresh()
		{
			// ファイル更新日時を取得
			SysTime modify = std.file.timeLastModified(path);

			// 更新されていなければ何もしない
			if (time == modify.stdTime())
				return;

			// ファイルを読み込み
			string data;
			try
			{
				data = cast(string)std.file.read(path);
			}
			catch (FileException fe)
			{
				return;
			}

			// ファイルをパースしてセット
			auto parser = new SweetsParser;
			elements = parser.parse(data).elements;

			// ファイル更新日時を保存
			time = modify.stdTime();

			// リビジョンを更新
			++revision;
		}	
	}

	// ファイルのマップ
	shared static
	File[string] files_;

	// 使用するファイル
	File file_;

	// テンプレートデータ
	SweetsElement[] elements_;

	// リビジョン番号
	uint revision_;

	/+
	 + コンストラクタ
	 +/
	this(File file)
	{
		file_ = file;
	}
}

/++
 + メモリ上にプリントするプリンタ
 +/
class MemoryPrinter
{
	/++
	 + バッファへの参照を取得
	 +
	 + Returns:
	 +  バッファへの参照
	 +/
	const
	const(char)[] buffer()
	nothrow
	{
		return buffer_[0 .. write_];
	}

	/++
	 + 確保したバッファは解放せずに内容だけクリアする
	 +/
	void clear()
	nothrow
	{
		write_ = 0;
	}

	/++
	 + 保持しているバッファに受け取ったデータを書き足す
	 +
	 + Params:
	 +  data = 追記するデータ
	 +/
	void print(const(char)[] data)
	nothrow
	{
		while (buffer_.length < write_ + data.length)
			buffer_.length = buffer_.length * 2 + 1;

		char* write = &buffer_[write_];

		foreach (d; data)
			*write++ = d;

		write_ += data.length;
	}

private:

	// 書き込みカーソル
	uint write_;

	// 書き込みバッファ
	char[] buffer_;
}

/++
 + ループ処理
 +
 + 基本的には配列と同じに使う
 + 追加で2つの機能がある
 + processメソッドをSweetsプロセッサとして登録可
 + currentでループ中のオブジェクトを取得
 +/
struct ForeachProcessor (T)
{
	// カレント要素
	@property
	T current()
	{
		return array[index_];
	}

	// 要素のインデックス
	@property
	uint index()
	{
		return index_;
	}

	// 配列実体
	T[] array;

	/++
	 + 子のみをループさせる
	 +
	 + Params:
	 +  element = 処理対象の要素
	 +  printer = 出力に使用するdelegate
	 +/
	void process(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		const(string)* f = "s:foreach" in element.attributes;
		if (f !is null && std.string.cmp(*f, "together") == 0)
		{
			for (index_ = 0; index_ < array.length; ++index_)
			{
				element.printStartTag(printer, false);
				foreach (child; children)
					child.print(printer);
				element.printEndTag(printer);
			}
		}
		else if (f !is null && std.string.cmp(*f, "insideonly") == 0)
		{
			for (index_ = 0; index_ < array.length; ++index_)
			{
				foreach (child; children)
					child.print(printer);
			}
		}
		else
		{
			element.printStartTag(printer, false);
			for (index_ = 0; index_ < array.length; ++index_)
			{
				foreach (child; children)
					child.print(printer);
			}
			element.printEndTag(printer);
		}
	}

	/++
	 + 代入のショートカット
	 +
	 + 普通に代入するだけならば配列に見えるようにするショートカット
	 +
	 + Params:
	 +  array = 代入する値
	 +
	 + Retusn:
	 +  代入した値
	 +/
	T[] opAssign(T[] array)
	{
		this.array = array;
		return array;
	}

private:

	// 処理中のインデックス
	uint index_;
}
unittest
{
	char[] buffer;
	void printer(const(char)[] data)
	{
		buffer ~= data;
	}

	auto parser = new SweetsParser;

	// 3回ループをセット
	auto processor = new ForeachProcessor!(int);
	processor.array = [0, 1, 2];

	// テンプレート
	string[] templates =
	[
		"<div s:id=\"hoge\">hogehoge</div>",
		"<div s:id=\"hoge\" s:foreach=\"together\">hogehoge</div>",
		"<div s:id=\"hoge\" s:foreach=\"insideonly\">hogehoge</div>",
	];

	// 想定する処理結果
	string[] results =
	[
		"<div>hogehogehogehogehogehoge</div>",
		"<div>hogehoge</div><div>hogehoge</div><div>hogehoge</div>",
		"hogehogehogehogehogehoge",
	];

	// 処理を実行して想定の結果と比較する
	for (uint i; i < templates.length; ++i)
	{
		auto tpl = parser.parse(templates[i]);
		auto sweets = new Sweets(tpl);
		sweets.assign("hoge", &processor.process);

		buffer = [];
		sweets.print(&printer);
		assert(std.string.cmp(results[i], buffer.idup) == 0);
	}
}
